# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import json

import pkg_resources

from . import utils
from . import branches
from . import instances
from . import frontend
from . import gitlab


class TestsManager:
    """Main instance of the application"""

    def __init__(self, root, config):
        self.config = config
        self.root = os.path.abspath(root)

        self.gitlab = gitlab.GitLabAPI(self)

        hooks_processor = frontend.HooksProcessor(self)
        self.frontend = frontend.create_app(self, hooks_processor)
        self.instances = instances.InstancesManager(self, self.frontend,
                                                    hooks_processor)

        if not utils.is_root_valid(root):
            raise RuntimeError("Invalid root directory: %s" % root)

        # Shortcut for creating multiple attributes
        dirs = ["git", "envs", "caches", "public", "socks", "build"]
        for dir in dirs:
            setattr(self, "%s_dir" % dir, os.path.join(root, dir))

        self.config_file = os.path.join(root, "config.json")
        self.gunicorn_config_file = pkg_resources.resource_filename(
            "managetests", "gunicorn_config"
        )

        self._load_details()
        self._load_branches()

    def _load_details(self):
        """Load details from the root directory"""
        with open(os.path.join(self.root, "details.json")) as f:
            self.details = json.load(f)

    def _load_branches(self):
        """Load all the branches"""
        self.branches = {}
        for branch_name, mr in self.details["branches"].copy().items():
            self.load_branch(branch_name, mr)

        # Load also new forced-to-be-keeped branches
        for branch_name in self.config["keep-branches"]:
            if branch_name in self.branches:
                continue
            self.load_branch(branch_name, None)

        # Load also new branches from merge requests
        new = self.gitlab.merge_requests()
        if new is None:
            raise RuntimeError("Can't get merge requests from GitLab!")
        for request in new:
            # Skip already loaded requests
            if request["source_branch"] in self.branches:
                continue
            self.load_branch(request["source_branch"], request["id"])

    def load_branch(self, name, mr):
        """Load a branch"""
        branch = branches.Branch(self, name, mr)
        if not branch.present and branch.active:
            branch.deploy()
        elif not branch.active:
            if branch.present:
                branch.destroy()
            return
        self.branches[name] = branch

        # Load the branch in the runner
        self.instances.load_branch(branch)

        # Save the branch details
        self.details["branches"][name] = mr
        self.save_details()

    def remove_branch(self, name):
        """Remove a branch"""
        if name not in self.branches:
            return

        self.instances.remove_branch(name)

        self.branches[name].destroy()
        del self.branches[name]

        del self.details["branches"][name]
        self.save_details()

    def save_details(self):
        """Save details on disk"""
        with open(os.path.join(self.root, "details.json"), "w") as f:
            json.dump(self.details, f)

    def run(self):
        """Run all the test instances"""
        self.instances.run()

    def process_hook(self, data):
        """Process a webhook received by GitLab"""
        obj = data["object_attributes"]
        branch = obj["source_branch"]

        # The branch was loaded
        if branch in self.branches:
            # Delete old branches
            if obj["state"] != "opened":
                self.remove_branch(branch)
                self.gitlab.post_comment(obj["id"], "L'istanza è stata rimossa.")
                return

            # Get the local commit
            with open("%s/%s/commit" % (self.envs_dir, branch)) as f:
                commit = f.read()

            # Rebuild if there is a new commit
            if commit != obj["last_commit"]["id"]:
                self.remove_branch(branch)
                self.load_branch(branch, obj["id"])
                self.gitlab.post_comment(obj["id"], "L'istanza è stata "
                    "aggiornata con il nuovo codice.")

        # The instance should be created
        elif obj["state"] == "opened":
            self.load_branch(branch, obj["id"])
            self.gitlab.post_comment(obj["id"], "Istanza live disponibile su "
                "http://wwwtest.ubuntu-it.org/%s." % branch)
